/* emmimp.h, EMM import structure
 * copyright (c) taQ 1994
 */

typedef unsigned long uint32;	/* double word */
typedef unsigned int uint16;	/* word */
typedef unsigned char uint8;	/* byte */

/* Windows/emm386 EMM import structure definition. The EMM import strucuture
 * consists of:
 * 1. hdr (fixed size, includes wems_frame_stat_t for each 16K frame in 1st 1MB)
 * 2. UMB frame descriptors
 * 3. EMS handle info
 * 4. other info (int 67 trap, HMA/XMS/UMB free area info, free mem, ...)
 * The size of the structure is approx. 0x200 bytes if no EMS handles have been
 * allocated.
 *
 * A separate pagemap that shows which EMS handle owns which physical pages
 * exists elsewhere.
 *
 * There are 3 versions of the structure, which can be identified by the 
 * "version" entry in the hdr. 
 * version 0x100: contains only EMS info. Can be used with both Win 3.0 & 3.1
 * version 0x10a: contains UMB/XMS/HMA/EMS info. Can be used only with Win 3.1
 * version 0x10b: contains same as 0x10a, but also contains maker/product name
 *                of the memory manager.
 * Therefore, most memory manager return ver 0x100 structure with Win 3.0,
 * 0x10b with Win 3.1
 */

 /* all the structures below are "packed" */

/*---------------------------------------------------------------------------
 * typedefs before typedef'ing the header (wems_hdr_t)
 *--------------------------------------------------------------------------*/

typedef struct {
	uint8 type;	/* b0: = 0 : non-EMSframe, = 1: EMS frame
			 * b1: if EMS, = 0: large-frame, = 1: non-large
			 * if frame includes UMB, set b2 - b5 to 1
			 * (b2: = 1 if first 4k is UMB,
			 *  b3: = 1 if second 4k is UMB,...)
			 */
	uint8 handle;	/* owner handle, 0xff or 0x00 for none,
			 * for frame including UMB: index to UMB frame
			 * descriptors
			 */
	uint16 logpage;	/* non-EMS frame: 0xffff
			 * EMSframe: logical page
			 *     if (handle != 0xff & handle != 0x00)
			 *         then 0x7fff for none
			 *     else
			 *         don't care
			 */
	uint8 physpg_num;	/* EMS physical page #
				 * e.g. if EMS frame at E000h, then
				 * 0 for frame at e000, 1 for e400, ...
				 * 0xff for non EMS page (don't care?) */
	uint8 flag;	/* EMS frame: 0
			 * non-EMS frame: 2 bits for each phys page.
			 * (i.e. b0-b1: first 4k, b2-b3: second 4k, ...)
			 * 10: direct mapping (physical == linear)
			 * 01: UMB mapping (physical != linear)
			 * corresponds to b2 - b5 of type field
			 * 386max always set this to 0
			 */
} wems_framestat_t;	/* for each frame (16KB), 6 bytes */


/*---------------------------------------------------------------------------
 * The real thing (pointed to by EMM import struct record)
 *--------------------------------------------------------------------------*/

/* wems_hdr_t definition */

typedef struct {
	uint8  flag0;	/* b3 (0x08) = 1: free emm386 virtual HMA 
			 *   only if hma_page_table_paddr != 0
			 *   hma_page_table_paddr is in other-info. (below)
			 * b4 (0x10) = 1: no UMB?.
			 * will have same effect if num_umb_pg_desc == 0
			 * other bits are unknown. normally set to 0x04
			 */

	uint8  resv0;		/* set this to 0 */

	uint16 size;		/* valid size of this info, from top of
				 * hdr. (in bytes) */

	uint16 version;		/* set to 0x100: no UMB, 0x10a: has UMB,
				 * 0x10b:has UMB + maker/product */
	uint32 resv1;		/* set to 0, Windows accesses this as uint32 */
/* offset 0ah */
	wems_framestat_t frame[64];	/* each 6 bytes, 64 entries */
/* offset 018ah */
	unsigned char resv2;	/* must be >= (# of EMS frames) * 3 + 4
				 * (PC-9801).
				 * PC-AT memory managers sets this to a much
				 * larger value.
				 */
/* offset 018bh */

 } wems_hdr_t;

/* after wems_hdr_t:
 * offset 0 (byte): num_umb_frame_desc
 *	# of UMB frame descriptors.
 *	each UMB frame descriptor is:
 *	physpg# of first 4k, physpg# of 2nd 4k, ...
 *		(16 bytes)
 *	N.B. NOT pagetbl-entry value.
 *	if non-UMB physpage in frame set that physpage to 0
 *	for version before 0x100, set this to 0
 * offset 1 - ...:
 *  an array of wems_umb_frame_desc_t (see below)
 */

typedef struct {
	uint32 physpgnum[4];	/* 4 phys-page#s. if non-UMB physpage, set to 0 */
} wems_umb_frame_desc_t;

/* after umb_frame_desc_t, 
 * offset 0:
 *	num_handle_info (1 byte) 
 * offset 1 - ...:
 *	an array of wems_handle_info_t (see below)
 * 
 * There always is an entry for system handle.
 * therefore, num_handles == 1 + number of non-system handles
 * 	(when Windows starts up.)
 */

typedef struct {
	uint8 handle;	/* handle #, 0 for system handle */
	uint8 flag;	/* b0 = 0 for handle #0,
			 *	1 for normal handle. 
			 *	1 even if pages for handle == 0
			 * may be Not Used (386 max does not set this)
			 */
	unsigned char hndlname[8];	/* name assigned to EMS handle
			 		 * (via AX = 5301h/int 67h) */
	uint16 pages;	/* in EMS (16 KB) pages */
	uint32 pgmap_phys_addr;	/* page map physical address,
				 * see below */
} wems_handle_info_t;

/* N.B. set all values to 0 for handle 0 (system handle), if no large frame
 * (i.e. frames in 640K) exists. If large frames exist, then entries
 * "pages", "pgmap_phys_addr" & a corresponding pagemap must be set.
 */
/* pagemap (pointed to by pgmap_phys_addr) contains i386 page-table values
 * for each page.
 * therefore, for each EMS page, we need 4 entries, each 4 bytes long.
 * (e.g. for a EMS handle with 2 pages, first page mapped to phys addr 0x130000,
 *  second page to 0x140000, the entries at pgmap_phys_addr will be:
 * 	0x130067, 0x131067, 0x132067, 0x133067,
 * 	0x140067, 0x141067, 0x142067, 0x143067
 *		where 67: bits for i386 page table entry.
 */

/* version 0x100 ends here */

/* after wems_handle_info_t are other info.
 * offset 0 (dword):	(int67 realmode vector)
 *    realmode int67 vector value (seg:ofs pair, 4 bytes),
 *    used for setting breakpoints (by Windows)
 *
 *  offset 4 (dword): (hma_page_table_paddr)
 *    phys addr of hma_page_table (page-table values for HMA, 64 bytes)
 *    if non-0, then
 *	a. if bit3 of wems_hdr_t.flag0 == 1 and page-table does not contain
 *	   pages below 1M+64K and all pages are present,
 *	   then all pages are _AddFreePhysPage'ed
 *      b. if bit3 ... == 0, then the page-table is  Set_Physical_HMA_Alias'ed
 *    (From NEC PC-9801 V86MMGR)
 *
 * offset 8 (byte): (free_pg_entry_count)
 *	# of free_page_entries[], which follows immediately after.
 *	may be 0.
 *
 * offset 9 - (free_pg_entry_count * sizeof free_pg_entry): free_pg_entries[]
 *	list of phys pages that Windows can use (via _AddFreePhysPage).
 */

typedef struct {
	uint32 physpgnum; /* physical pg #, not PTE val. */
	uint32 numpg;	  /* # of physical pages starting from physpgnum */
} free_pg_entry_t;

/* offset M (byte): # of XMS handle info (each size 0xc bytes)
 *		0 if memory manager does not emulate XMS, 
 *		or has realmode XMS code that can be executed in
 *		Windows-environment
 *
 * offset M + 1: array of XMS handle info. see wems_xms_info_t (below) 
 *
 * offset N (byte): (umb_free_info_count)
 *	# of umb_free_info[], which follows immediately after. may be 0
 *
 * offset N + 1 - ...: (umb_free_info[])
 *	info for free UMB mem arena.
 */
typedef struct {
	uint16 startseg;	/* start of free UMB arena */
	uint16 parasize;	/* size in paragraph (16 bytes) */
}  umb_free_seg_t;

/*	startseg is start address of arena (in realmode segment).
 *	note that this is NOT the segment of arena-header ('M' or 'Z'), 
 *	but + 1.
 *	parasize is size of that arena, which does not include the arena-header.
 *	e.g. if arena header is (at c0bf:0)
 *	C0BF:0000  4D 00 00 3F 1B FF FF FF-00 00 00 00 00 00 00 00
 *	then startseg = 0xc0c0, parasize = 0x1b3f
 *
 *	if "DOS = UMB" is defined in config.sys, then umb_free_info_count would
 *	be 0.
 *
 * version 0x10a ends here.
 *
 * offset NN: maker/product name (version 0x10b)
 *	maker name is the name of the maker of this mem manager
 *	(e.g. "MICROSOFT"), 0x14 bytes, ' ' padded.
 * 	product name immediately follows maker name,  0x14 bytes, ' ' padded.
 *	(e.g. "EMM386 4.44")
 *
 * version 0x10b ends here.
 *
 */

typedef struct {
	uint16 handle;
	uint16 flag;	/* b0 = 0: handles already in use when
			 * Windows starts up
			 * b0 = 1: handles usable by Windows
			 *      (can provide to DOS-Prompt) */
	/* if flag.bit0 == 0, fields below are used, else are ignored */
	uint32 kb; 	/* size in KB (may be 0) */
	uint32 baseaddr;	/* physical address of this handle */
 } wems_xms_info_t;

 /* N.B. do not put info for handles that Windows itself has allocated.
  * Assume that all XMS allocation is done by Windows after ax = 1605h/int 2fh,
  * and before ax = 1606h/int 2fh
  * if there are any handles with flag.b0 == 1, Windows will use these
  * handles when programs running in DOS-Prompt allocate XMS memory.
  * 
  * if there are no handles with flag.b0 == 1, then Windows
  * calls realmode XMS code (ah = 9/dx = 0, i.e. allocate emb, with size == 0) 
  * for handles.
  */


/* usage: physical address to wems_hdr_t must be returned on
 * ioctl-read to EMMXXXX0, but only when ioctl-buffer[0] == 1, read length 6.
 * (ioctl-read occurs during realmode initialization of vxd V86MMGR, 
 * which is afer ax = 1605h/int 2fh)
 * address is returned into buffer[0:3], buffer[4] <- 1, buffer[5] <- 0 ~ 0xb
 * (see v86mmgr.vxd, object 2 (realmode init code). this phys addr
 *   is then passed to protmode code as reference data)
 * wems_hdr_t must be dword aligned? (386Max does not)
 */
