/*
** File:        tga221.c             Copyright (c) Truda Software
** Author:      Anton Kruger         215 Marengo Rd, #2, 
** Date:        March 1992           Oxford, IA 52322-9383
** Revision:    1.0                  March 1992
** 
** Description: A driver for "median". Reads in Targa type 2
**              (true-color) image files, applies the median-cut
**              algorithm to find 256 good colors for the image,
**              and writes out Targa type 1 (color-mapped) file.
**
** Compilers:  MSC 5.1, 6.0.
**
** Note:       Compile in large memory model.
*/
#include <stdlib.h>
#include <stdio.h>
#define IOBUFFSIZE 8192          /* size of i/o buffers        */
#define MAXCOLORS  256           /* maximum # of output colors */
#define HSIZE      32768         /* size of image histogram    */
typedef unsigned char byte;      /* range 0-255                */
typedef unsigned short word;     /* range 0-65,535             */
typedef unsigned long dword;     /* range 0-4,294,967,295      */

/* Macros for converting between (r,g,b)-colors and 15-bit     */
/* colors follow.                                              */
#define RGB(r,g,b) (word)(((b)&~7)<<7)|(((g)&~7)<<2)|((r)>>3)
#define RED(x)     (byte)(((x)&31)<<3)
#define GREEN(x)   (byte)((((x)>>5)&255)<< 3)
#define BLUE(x)    (byte)((((x)>>10)&255)<< 3)

word Hist[HSIZE];                /* 15-bit image histogram     */

struct ImageHeader {   /* Offset          Meaning              */ 
   byte  IdSize;       /*  0   id field size at end of header  */
   byte  ColMapType;   /*  1   0=>no color map, 1=>color map   */
   byte  TypeCode;     /*  2   1=>type 1, 2=>type 2 images,... */
   word  ColMapOrigin; /*  3   index of first color map entry  */
   word  ColMapLength; /*  5   number of entries in color map  */
   byte  ColMapBits;   /*  7   number of bits/color map entry  */
   word  XOrigin;      /*  8   lower left corner of image      */
   word  YOrigin;      /*  10  lower left corner of image      */
   word  Width;        /*  12  width of image in pixels        */
   word  Height;       /*  14  height of image in pixels       */
   byte  BitsPerPixel; /*  16  see note 1: below               */
   byte  Descriptor;   /*  17  see note 2: below               */
   byte  IdField[256]; /*  18  image identification field      */
   /*                                                          */
   /* Note 1: The "BitsPerPixel" field depends of the type of  */
   /*         Targa file. For Targa type 1 images, it is the   */
   /*         number of bits per pixel index.  For a true-     */
   /*         color Targa image this is 24, since pixels need  */
   /*         1 byte each for r, g, and b.                     */
   /* Note 2: Individual bits have meaning here. For example,  */
   /*         toggling bit 5 flips the image upside down.      */
};


void ReadTargaHeader(FILE *fp1, struct ImageHeader *Image);
void WriteTargaHeader(FILE *fp1, struct ImageHeader Image);
void getrgb(byte * r,byte * g,byte * b, FILE *fp1);
word MedianCut(word Hist[],byte ColMap[][3], int maxcubes);
void inform(char *mess);
void main(int argc, char *argv[]);


void main(int argc, char *argv[])
{
   FILE               *fp1,*fp2;
   byte               r,g,b;
   byte               index,ColMap[256][3];
   word               i,n,color;
   struct ImageHeader Image;

   if (argc < 3){
      inform("Usage: MEDIAN infile outfile\n\n");
      inform("infile is a true-color Targa type 2 file\n");
      inform("outfile is a color-mapped Targa type 1 file\n");
      exit(-1);
   }
   if ((fp1 = fopen(argv[1],"rb")) == NULL) {
      inform("ERROR: Could not open input file: ");
      inform(argv[1]);
      exit(-1);
   }
   setvbuf(fp1,(char *)NULL,_IOFBF,IOBUFFSIZE);
   if ((fp2 = fopen(argv[2],"wb")) == NULL) {
      inform("ERROR: Could not open output file: ");
      inform(argv[2]);
      exit(-1);
   }
   setvbuf(fp2,(char *)NULL,_IOFBF,IOBUFFSIZE);
   inform("Median-Cut Color Quantization.\n");
   inform(argv[1]); inform(" -> "); inform(argv[2]);
   inform("\nPlease wait...\n");

   /*
   ** Clear the histogram, read the image header and then build
   ** the image histogram. Next call "MedianCut" to quantize.
   */

   i = (word)HSIZE;
   while (i--)
      Hist[i] = (word)0;

   ReadTargaHeader(fp1, &Image);

   getrgb(&r,&g,&b,fp1);
   while (!feof(fp1)) {
      color = RGB(r,g,b);
      Hist[color] = Hist[color] + 1;
      getrgb(&r,&g,&b,fp1);
   }
   n = MedianCut(Hist,ColMap,(int)MAXCOLORS);

   /*
   ** Rewind the input file, and skip over the header. We are
   ** going write out a Targa type 1 image, so make the proper
   ** changes to the image header. Then write it out.
   */

   rewind(fp1);
   ReadTargaHeader(fp1,&Image);
   Image.ColMapType   = (byte)1;  /* type 1 image              */
   Image.TypeCode     = (byte)1;  /* yes there is a color map  */
   Image.ColMapOrigin = (word)0;  /* use color map all entries */
   Image.ColMapLength = n;        /* number of output colors   */
   Image.ColMapBits   = (byte)24; /* 1 byte/color in color map */
   Image.BitsPerPixel = (byte)8;  /* pixel index 8 bits wide   */
   WriteTargaHeader(fp2,Image);

   /* Now write out the color map, 1 byte/color */
   
   for (i=0;i<=n-1;i++){
      b = ColMap[i][2]; fputc(b,fp2); 
      g = ColMap[i][1]; fputc(g,fp2);
      r = ColMap[i][0]; fputc(r,fp2);
   }

   /*
   ** Finally, we can remap the input image. For each pixel in
   ** the input image, convert pixel's color to a 15-bit color.
   ** The pixel's index in the color map is looked up in "Hist"
   ** which now functions as an inverse color map.
   */

   getrgb(&r,&g,&b,fp1);
   while (!feof(fp1)) {
      color = RGB(r,g,b);           /* convert to 15-bit color */
      index = (byte)Hist[color];    /* index in the color map  */
      fputc(index,fp2);             /* write out pixel' index  */
      getrgb(&r,&g,&b,fp1);      
   }
   fclose(fp1);
   fclose(fp2);
}


void getrgb(byte * r,byte * g,byte * b, FILE *fp1)
{
   *b = (byte)getc(fp1); 
   *g = (byte)getc(fp1);
   *r = (byte)getc(fp1);
   return;
}


void WriteTargaHeader(FILE *fp1, struct ImageHeader Image)
{
   /*  
   ** Writes the Targa header information to a file. Note the
   ** two macros below work reliable ONLY for 2-byte variables.
   ** "LSB" returns the lower byte, "MSB" returns the upper byte.
   */
#define LSB(x)  ((x) & 255)
#define MSB(x)  (((x)>>8)&255)
   
   byte        header[18];
   size_t      nbytes;

   header[0]  = (byte)Image.IdSize;
   header[1]  = (byte)Image.ColMapType;
   header[2]  = (byte)Image.TypeCode;
   header[3]  = (byte)LSB(Image.ColMapOrigin);
   header[4]  = (byte)MSB(Image.ColMapOrigin);
   header[5]  = (byte)LSB(Image.ColMapLength);
   header[6]  = (byte)MSB(Image.ColMapLength);
   header[7]  = (byte)Image.ColMapBits;
   header[8]  = (byte)LSB(Image.XOrigin);
   header[9]  = (byte)MSB(Image.XOrigin);
   header[10] = (byte)LSB(Image.YOrigin);
   header[11] = (byte)MSB(Image.YOrigin);
   header[12] = (byte)LSB(Image.Width);
   header[13] = (byte)MSB(Image.Width);
   header[14] = (byte)LSB(Image.Height);
   header[15] = (byte)MSB(Image.Height);
   header[16] = (byte)Image.BitsPerPixel;
   header[17] = (byte)Image.Descriptor;
   nbytes = 18;
   fwrite((void *)header,sizeof(byte),nbytes,fp1);

   /* Write out image identification field, if any */

   if ((nbytes = Image.IdSize) != 0)
      fwrite((void *)Image.IdField,sizeof(byte),nbytes,fp1);
}


void ReadTargaHeader(FILE *fp1, struct ImageHeader *Image)
{
   /*
   ** Reads a Targa file header, and fills "Image" structure.
   */

   byte        header[18];
   size_t      nbytes;

   nbytes = 18;
   if (fread((void *)header,sizeof(byte),nbytes,fp1) != nbytes){
      inform("ERROR: Unexpected EOF\n");
      exit(-1);
   }
   Image->IdSize       = header[0];
   Image->ColMapType   = header[1];
   Image->TypeCode     = header[2];
   Image->ColMapOrigin = header[3] + 256*header[4];
   Image->ColMapLength = header[5] + 256*header[6];
   Image->ColMapBits   = header[7];
   Image->XOrigin      = header[8]  + 256*header[9];
   Image->YOrigin      = header[10] + 256*header[11];
   Image->Width        = header[12] + 256*header[13];
   Image->Height       = header[14] + 256*header[15]; 
   Image->BitsPerPixel = header[16];
   Image->Descriptor   = header[17];

   /* Read in image identification field, if any */

   if ((nbytes = Image->IdSize) != 0)
      fread((void*)Image->IdField,sizeof(byte),nbytes,fp1);
}


void inform(char *mess)
{
   fprintf(stderr,"%s",mess);
}
